
// #include "models.h"
#include "includes.h"
#include "auth.h"
#include "librpc/gen_ndr/ioctl.h"
#include "smbd/smbd.h"
#include "substitute.h"
#include "system/dir.h"
#include "system/filesys.h"

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <time.h>
#include <unistd.h>

// static const mode_t FIFO_MODE = 0600;

#define SS_DEFAULT_SOCKET_PATH "/var/run/snapshield/fs_event.sock"

#define SS_DEBUG_SOCKET_FAIL_LOG_LEVEL 1

#define SS_REPORT_OPENAT_READS 1
#define SS_VERBOSE_OPENAT_FLAGS 0
#define SS_LOG_EUID 0

#define SS_MESSAGE_HEADER_MAGIC_NUMBER 0x4545454545454545

#define DEBUG_ALL_MESSAGES

typedef struct ss_conn_info {
	const char *ip_address;
	const char *user;
	const char *remote_machine_name;
	const char *share_name;
} ss_conn_info;

typedef struct ss_socket_info {
	const char *socket_path;
	int socket_connection_fd;
} ss_socket_info;

typedef struct ss_private_data {
	ss_conn_info conn_info;
	ss_socket_info socket_info;
} ss_private_data;

// never change
typedef enum _vfs_op_type {
	SMB_VFS_OP_NOOP = -1,

	/* Disk operations */

	SMB_VFS_OP_CONNECT = 0,
	SMB_VFS_OP_DISCONNECT,
	SMB_VFS_OP_DISK_FREE,
	SMB_VFS_OP_GET_QUOTA,
	SMB_VFS_OP_SET_QUOTA,
	SMB_VFS_OP_GET_SHADOW_COPY_DATA,
	SMB_VFS_OP_STATVFS,
	SMB_VFS_OP_FS_CAPABILITIES,
	SMB_VFS_OP_GET_DFS_REFERRALS,
	SMB_VFS_OP_CREATE_DFS_PATHAT,
	SMB_VFS_OP_READ_DFS_PATHAT,

	/* Directory operations */

	SMB_VFS_OP_FDOPENDIR,
	SMB_VFS_OP_READDIR,
	SMB_VFS_OP_SEEKDIR,
	SMB_VFS_OP_TELLDIR,
	SMB_VFS_OP_REWINDDIR,
	SMB_VFS_OP_MKDIRAT,
	SMB_VFS_OP_CLOSEDIR,

	/* File operations */

	SMB_VFS_OP_OPEN,
	SMB_VFS_OP_OPENAT,
	SMB_VFS_OP_CREATE_FILE,
	SMB_VFS_OP_CLOSE,
	SMB_VFS_OP_READ,
	SMB_VFS_OP_PREAD,
	SMB_VFS_OP_PREAD_SEND,
	SMB_VFS_OP_PREAD_RECV,
	SMB_VFS_OP_WRITE,
	SMB_VFS_OP_PWRITE,
	SMB_VFS_OP_PWRITE_SEND,
	SMB_VFS_OP_PWRITE_RECV,
	SMB_VFS_OP_LSEEK,
	SMB_VFS_OP_SENDFILE,
	SMB_VFS_OP_RECVFILE,
	SMB_VFS_OP_RENAMEAT,
	SMB_VFS_OP_FSYNC,
	SMB_VFS_OP_FSYNC_SEND,
	SMB_VFS_OP_FSYNC_RECV,
	SMB_VFS_OP_STAT,
	SMB_VFS_OP_FSTAT,
	SMB_VFS_OP_LSTAT,
	SMB_VFS_OP_GET_ALLOC_SIZE,
	SMB_VFS_OP_UNLINKAT,
	SMB_VFS_OP_FCHMOD,
	SMB_VFS_OP_FCHOWN,
	SMB_VFS_OP_LCHOWN,
	SMB_VFS_OP_CHDIR,
	SMB_VFS_OP_GETWD,
	SMB_VFS_OP_NTIMES,
	SMB_VFS_OP_FNTIMES,
	SMB_VFS_OP_FTRUNCATE,
	SMB_VFS_OP_FALLOCATE,
	SMB_VFS_OP_LOCK,
	SMB_VFS_OP_KERNEL_FLOCK,
	SMB_VFS_OP_FCNTL,
	SMB_VFS_OP_LINUX_SETLEASE,
	SMB_VFS_OP_GETLOCK,
	SMB_VFS_OP_SYMLINKAT,
	SMB_VFS_OP_READLINKAT,
	SMB_VFS_OP_LINKAT,
	SMB_VFS_OP_MKNODAT,
	SMB_VFS_OP_REALPATH,
	SMB_VFS_OP_FCHFLAGS,
	SMB_VFS_OP_FILE_ID_CREATE,
	SMB_VFS_OP_FS_FILE_ID,
	SMB_VFS_OP_FSTREAMINFO,
	SMB_VFS_OP_GET_REAL_FILENAME,
	SMB_VFS_OP_CONNECTPATH,
	SMB_VFS_OP_BRL_LOCK_WINDOWS,
	SMB_VFS_OP_BRL_UNLOCK_WINDOWS,
	SMB_VFS_OP_STRICT_LOCK_CHECK,
	SMB_VFS_OP_TRANSLATE_NAME,
	SMB_VFS_OP_PARENT_PATHNAME,
	SMB_VFS_OP_FSCTL,
	SMB_VFS_OP_OFFLOAD_READ_SEND,
	SMB_VFS_OP_OFFLOAD_READ_RECV,
	SMB_VFS_OP_OFFLOAD_WRITE_SEND,
	SMB_VFS_OP_OFFLOAD_WRITE_RECV,
	SMB_VFS_OP_FGET_COMPRESSION,
	SMB_VFS_OP_SET_COMPRESSION,
	SMB_VFS_OP_SNAP_CHECK_PATH,
	SMB_VFS_OP_SNAP_CREATE,
	SMB_VFS_OP_SNAP_DELETE,

	/* DOS attribute operations. */
	SMB_VFS_OP_GET_DOS_ATTRIBUTES_SEND,
	SMB_VFS_OP_GET_DOS_ATTRIBUTES_RECV,
	SMB_VFS_OP_FGET_DOS_ATTRIBUTES,
	SMB_VFS_OP_FSET_DOS_ATTRIBUTES,

	/* NT ACL operations. */

	SMB_VFS_OP_FGET_NT_ACL,
	SMB_VFS_OP_FSET_NT_ACL,
	SMB_VFS_OP_AUDIT_FILE,

	/* POSIX ACL operations. */

	SMB_VFS_OP_SYS_ACL_GET_FD,
	SMB_VFS_OP_SYS_ACL_BLOB_GET_FD,
	SMB_VFS_OP_SYS_ACL_SET_FD,
	SMB_VFS_OP_SYS_ACL_DELETE_DEF_FD,

	/* EA operations. */
	SMB_VFS_OP_GETXATTRAT_SEND,
	SMB_VFS_OP_GETXATTRAT_RECV,
	SMB_VFS_OP_FGETXATTR,
	SMB_VFS_OP_FLISTXATTR,
	SMB_VFS_OP_REMOVEXATTR,
	SMB_VFS_OP_FREMOVEXATTR,
	SMB_VFS_OP_FSETXATTR,

	/* aio operations */
	SMB_VFS_OP_AIO_FORCE,

	/* offline operations */
	SMB_VFS_OP_IS_OFFLINE,
	SMB_VFS_OP_SET_OFFLINE,

	/* Durable handle operations. */
	SMB_VFS_OP_DURABLE_COOKIE,
	SMB_VFS_OP_DURABLE_DISCONNECT,
	SMB_VFS_OP_DURABLE_RECONNECT,

	SMB_VFS_OP_FREADDIR_ATTR,

	/* This should always be last enum value */

	SMB_VFS_OP_LAST
} vfs_op_type;

/* The following array *must* be in the same order as defined in vfs_op_type */

static struct {
	vfs_op_type type;
	const char *name;
} vfs_op_names[] = {
	{SMB_VFS_OP_CONNECT, "connect"},
	{SMB_VFS_OP_DISCONNECT, "disconnect"},
	{SMB_VFS_OP_DISK_FREE, "disk_free"},
	{SMB_VFS_OP_GET_QUOTA, "get_quota"},
	{SMB_VFS_OP_SET_QUOTA, "set_quota"},
	{SMB_VFS_OP_GET_SHADOW_COPY_DATA, "get_shadow_copy_data"},
	{SMB_VFS_OP_STATVFS, "statvfs"},
	{SMB_VFS_OP_FS_CAPABILITIES, "fs_capabilities"},
	{SMB_VFS_OP_GET_DFS_REFERRALS, "get_dfs_referrals"},
	{SMB_VFS_OP_CREATE_DFS_PATHAT, "create_dfs_pathat"},
	{SMB_VFS_OP_READ_DFS_PATHAT, "read_dfs_pathat"},
	{SMB_VFS_OP_FDOPENDIR, "fdopendir"},
	{SMB_VFS_OP_READDIR, "readdir"},
	{SMB_VFS_OP_SEEKDIR, "seekdir"},
	{SMB_VFS_OP_TELLDIR, "telldir"},
	{SMB_VFS_OP_REWINDDIR, "rewinddir"},
	{SMB_VFS_OP_MKDIRAT, "mkdirat"},
	{SMB_VFS_OP_CLOSEDIR, "closedir"},
	{SMB_VFS_OP_OPEN, "open"},
	{SMB_VFS_OP_OPENAT, "openat"},
	{SMB_VFS_OP_CREATE_FILE, "create_file"},
	{SMB_VFS_OP_CLOSE, "close"},
	{SMB_VFS_OP_READ, "read"},
	{SMB_VFS_OP_PREAD, "pread"},
	{SMB_VFS_OP_PREAD_SEND, "pread_send"},
	{SMB_VFS_OP_PREAD_RECV, "pread_recv"},
	{SMB_VFS_OP_WRITE, "write"},
	{SMB_VFS_OP_PWRITE, "pwrite"},
	{SMB_VFS_OP_PWRITE_SEND, "pwrite_send"},
	{SMB_VFS_OP_PWRITE_RECV, "pwrite_recv"},
	{SMB_VFS_OP_LSEEK, "lseek"},
	{SMB_VFS_OP_SENDFILE, "sendfile"},
	{SMB_VFS_OP_RECVFILE, "recvfile"},
	{SMB_VFS_OP_RENAMEAT, "renameat"},
	{SMB_VFS_OP_FSYNC, "fsync"},
	{SMB_VFS_OP_FSYNC_SEND, "fsync_send"},
	{SMB_VFS_OP_FSYNC_RECV, "fsync_recv"},
	{SMB_VFS_OP_STAT, "stat"},
	{SMB_VFS_OP_FSTAT, "fstat"},
	{SMB_VFS_OP_LSTAT, "lstat"},
	{SMB_VFS_OP_GET_ALLOC_SIZE, "get_alloc_size"},
	{SMB_VFS_OP_UNLINKAT, "unlinkat"},
	{SMB_VFS_OP_FCHMOD, "fchmod"},
	{SMB_VFS_OP_FCHOWN, "fchown"},
	{SMB_VFS_OP_LCHOWN, "lchown"},
	{SMB_VFS_OP_CHDIR, "chdir"},
	{SMB_VFS_OP_GETWD, "getwd"},
	{SMB_VFS_OP_NTIMES, "ntimes"},
	{SMB_VFS_OP_FNTIMES, "fntimes"},
	{SMB_VFS_OP_FTRUNCATE, "ftruncate"},
	{SMB_VFS_OP_FALLOCATE, "fallocate"},
	{SMB_VFS_OP_LOCK, "lock"},
	{SMB_VFS_OP_KERNEL_FLOCK, "kernel_flock"},
	{SMB_VFS_OP_FCNTL, "fcntl"},
	{SMB_VFS_OP_LINUX_SETLEASE, "linux_setlease"},
	{SMB_VFS_OP_GETLOCK, "getlock"},
	{SMB_VFS_OP_SYMLINKAT, "symlinkat"},
	{SMB_VFS_OP_READLINKAT, "readlinkat"},
	{SMB_VFS_OP_LINKAT, "linkat"},
	{SMB_VFS_OP_MKNODAT, "mknodat"},
	{SMB_VFS_OP_REALPATH, "realpath"},
	{SMB_VFS_OP_FCHFLAGS, "fchflags"},
	{SMB_VFS_OP_FILE_ID_CREATE, "file_id_create"},
	{SMB_VFS_OP_FS_FILE_ID, "fs_file_id"},
	{SMB_VFS_OP_FSTREAMINFO, "fstreaminfo"},
	{SMB_VFS_OP_GET_REAL_FILENAME, "get_real_filename"},
	{SMB_VFS_OP_CONNECTPATH, "connectpath"},
	{SMB_VFS_OP_BRL_LOCK_WINDOWS, "brl_lock_windows"},
	{SMB_VFS_OP_BRL_UNLOCK_WINDOWS, "brl_unlock_windows"},
	{SMB_VFS_OP_STRICT_LOCK_CHECK, "strict_lock_check"},
	{SMB_VFS_OP_TRANSLATE_NAME, "translate_name"},
	{SMB_VFS_OP_PARENT_PATHNAME, "parent_pathname"},
	{SMB_VFS_OP_FSCTL, "fsctl"},
	{SMB_VFS_OP_OFFLOAD_READ_SEND, "offload_read_send"},
	{SMB_VFS_OP_OFFLOAD_READ_RECV, "offload_read_recv"},
	{SMB_VFS_OP_OFFLOAD_WRITE_SEND, "offload_write_send"},
	{SMB_VFS_OP_OFFLOAD_WRITE_RECV, "offload_write_recv"},
	{SMB_VFS_OP_FGET_COMPRESSION, "fget_compression"},
	{SMB_VFS_OP_SET_COMPRESSION, "set_compression"},
	{SMB_VFS_OP_SNAP_CHECK_PATH, "snap_check_path"},
	{SMB_VFS_OP_SNAP_CREATE, "snap_create"},
	{SMB_VFS_OP_SNAP_DELETE, "snap_delete"},
	{SMB_VFS_OP_GET_DOS_ATTRIBUTES_SEND, "get_dos_attributes_send"},
	{SMB_VFS_OP_GET_DOS_ATTRIBUTES_RECV, "get_dos_attributes_recv"},
	{SMB_VFS_OP_FGET_DOS_ATTRIBUTES, "fget_dos_attributes"},
	{SMB_VFS_OP_FSET_DOS_ATTRIBUTES, "fset_dos_attributes"},
	{SMB_VFS_OP_FGET_NT_ACL, "fget_nt_acl"},
	{SMB_VFS_OP_FSET_NT_ACL, "fset_nt_acl"},
	{SMB_VFS_OP_AUDIT_FILE, "audit_file"},
	{SMB_VFS_OP_SYS_ACL_GET_FD, "sys_acl_get_fd"},
	{SMB_VFS_OP_SYS_ACL_BLOB_GET_FD, "sys_acl_blob_get_fd"},
	{SMB_VFS_OP_SYS_ACL_SET_FD, "sys_acl_set_fd"},
	{SMB_VFS_OP_SYS_ACL_DELETE_DEF_FD, "sys_acl_delete_def_fd"},
	{SMB_VFS_OP_GETXATTRAT_SEND, "getxattrat_send"},
	{SMB_VFS_OP_GETXATTRAT_RECV, "getxattrat_recv"},
	{SMB_VFS_OP_FGETXATTR, "fgetxattr"},
	{SMB_VFS_OP_FLISTXATTR, "flistxattr"},
	{SMB_VFS_OP_REMOVEXATTR, "removexattr"},
	{SMB_VFS_OP_FREMOVEXATTR, "fremovexattr"},
	{SMB_VFS_OP_FSETXATTR, "fsetxattr"},
	{SMB_VFS_OP_AIO_FORCE, "aio_force"},
	{SMB_VFS_OP_IS_OFFLINE, "is_offline"},
	{SMB_VFS_OP_SET_OFFLINE, "set_offline"},
	{SMB_VFS_OP_DURABLE_COOKIE, "durable_cookie"},
	{SMB_VFS_OP_DURABLE_DISCONNECT, "durable_disconnect"},
	{SMB_VFS_OP_DURABLE_RECONNECT, "durable_reconnect"},
	{SMB_VFS_OP_FREADDIR_ATTR, "freaddir_attr"},
	{SMB_VFS_OP_LAST, NULL}};

typedef enum ss_message_type {
	SS_MESSAGE_TYPE_FileSystemEvent,
	SS_MESSAGE_TYPE_Error,
	SS_MESSAGE_TYPE_COUNT
} ss_message_type;

typedef struct __attribute__((__packed__)) ss_message_header {
	uint64_t magic_number;
	int64_t timestamp;
	uint64_t message_length;
	uint64_t message_type;
} ss_message_header;

typedef struct __attribute__((__packed__)) ss_fs_event_message {
	struct ss_message_header header;
	int32_t flags;
	int32_t succeeded;
	int32_t _errno;
	// offset into data
	uint32_t ip_addr_offset;
	uint32_t ip_addr_length;
	// offset into data
	uint32_t user_name_offset;
	uint32_t user_name_length;
	// offset into data
	uint32_t remote_machine_name_offset;
	uint32_t remote_machine_name_length;
	// offset into data
	uint32_t share_name_offset;
	uint32_t share_name_length;
	// offset into data
	uint32_t action_offset;
	uint32_t action_length;
	// offset into data
	uint32_t path_offset;
	uint32_t path_length;
	// offset into data
	uint32_t path2_offset;
	uint32_t path2_length;
	char data[];
} ss_fs_event_message;

#define SS_FS_EVENT_MESSAGE_SIZE(DATA_LEN) \
	(sizeof(ss_fs_event_message) + DATA_LEN)

static inline void __ss_log_euid(const char *func)
{
	uid_t uid = getuid();
	DEBUG(0, ("%s: Current EUID: %d\n", func, uid));
}
#if defined(SS_LOG_EUID) && (SS_LOG_EUID == 1)
#define ss_log_euid() __ss_log_euid(__func__)
#else
#define ss_log_euid() ((void)0)
#endif

static const char *ss_vfs_opname(vfs_op_type op)
{
	if (op >= SMB_VFS_OP_LAST)
		return "INVALID VFS OP";
	return vfs_op_names[op].name;
}

static int ss_open_socket_connection(const char *socket_path)
{
	int sockfd;
	struct sockaddr_un server_addr;

	DEBUG(1, ("Connecting to snapshield socket at %s\n", socket_path));

	sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sockfd < 0) {
		return -1;
	}

	memset(&server_addr, 0, sizeof(server_addr));
	server_addr.sun_family = AF_UNIX;
	strncpy(server_addr.sun_path,
		socket_path,
		sizeof(server_addr.sun_path) - 1);

	if (connect(sockfd,
		    (struct sockaddr *)&server_addr,
		    sizeof(server_addr)) < 0) {
		close(sockfd);
		return -1;
	}

	return sockfd;
}

static ssize_t
ss_write(ss_socket_info *socket_info, const void *buf, size_t len)
{
	ssize_t bytes_written;
	int err;

	if (!socket_info) {
		errno = EINVAL;
		return -1;
	}

	if (socket_info->socket_connection_fd < 0) {
		socket_info->socket_connection_fd =
			ss_open_socket_connection(socket_info->socket_path);
		if (socket_info->socket_connection_fd < 0) {
			return -1;
		}
	}

	bytes_written = write(socket_info->socket_connection_fd, buf, len);
	if (bytes_written < 0) {
		err = errno;
		if (err == EPIPE || err == ECONNRESET) {
			// Connection closed, attempt to reopen the connection
			// on next ss_write
			close(socket_info->socket_connection_fd);
			socket_info->socket_connection_fd = -1;
		}
		errno = err;
	}
	return bytes_written;
}

static ssize_t
ss_safe_write(ss_socket_info *socket_info, const void *buf, size_t len)
{
	ssize_t bytes_written;
	size_t iters = 0;

	while ((bytes_written = ss_write(socket_info, buf, len)) < 0 &&
	       iters++ <= 100) {
		if (!(errno == EINTR || errno == EPIPE ||
		      errno == ECONNRESET)) {
			// not interrupted or connection closed
			return -1; // Other error, return error
		} // else try again
	}
	return bytes_written;
}

static TALLOC_CTX *tmp_ss_event_log_ctx = NULL;
/*
 * Get us a temporary talloc context usable just for event logs
 */
static TALLOC_CTX *ss_event_log_ctx(void)
{
	if (tmp_ss_event_log_ctx == NULL) {
		tmp_ss_event_log_ctx =
			talloc_named_const(NULL, 0, "ss_event_log_ctx");
	}
	return tmp_ss_event_log_ctx;
}

static struct ss_message_header *
ss_message_alloc(TALLOC_CTX *ctx, ss_message_type type, size_t message_size)
{
	struct ss_message_header *message =
		(struct ss_message_header *)talloc_zero_size(ctx, message_size);
	if (!message) {
		errno = ENOMEM;
		return NULL;
	}
	message->magic_number = SS_MESSAGE_HEADER_MAGIC_NUMBER;
	message->timestamp = time(NULL);
	message->message_length = message_size - sizeof(ss_message_header);
	message->message_type = type;
	return message;
}

static struct ss_message_header *
ss_message_realloc(TALLOC_CTX *ctx,
		   struct ss_message_header *message,
		   size_t message_size)
{
	message = talloc_realloc_size(ctx, message, message_size);
	if (!message) {
		errno = ENOMEM;
		return NULL;
	}
	message->message_length = message_size - sizeof(ss_message_header);
	return message;
}

static int ss_message_send(vfs_handle_struct *handle,
			   struct ss_message_header *message)
{
	struct ss_private_data *pd = NULL;

	SMB_VFS_HANDLE_GET_DATA(handle, pd, struct ss_private_data, return -1;);

	if (ss_safe_write(&pd->socket_info,
			  (void *)message,
			  sizeof(struct ss_message_header) +
				  message->message_length) < 0) {
		return -1;
	}
	return 0;
}

static void ss_fs_event_message_log(struct ss_message_header *message)
{
	struct ss_fs_event_message *e = (struct ss_fs_event_message *)message;

	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("\ttype: fs event\n"
	       "\tflags: %d\n"
	       "\tstatus: %s\n",
	       e->flags,
	       e->succeeded ? "OK" : strerror(e->_errno)));
	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("\tip: %.*s\n",
	       e->ip_addr_length,
	       &(e->data[e->ip_addr_offset])));
	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("\tuser: %.*s\n",
	       e->user_name_length,
	       &(e->data[e->user_name_offset])));
	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("\tremote machine name: %.*s\n",
	       e->remote_machine_name_length,
	       &(e->data[e->remote_machine_name_offset])));
	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("\tshare: %.*s\n",
	       e->share_name_length,
	       &(e->data[e->share_name_offset])));
	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("\taction: %.*s\n",
	       e->action_length,
	       &(e->data[e->action_offset])));
	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("\tpath: %.*s\n", e->path_length, &(e->data[e->path_offset])));
	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("\tpath2: %.*s\n",
	       e->path2_length,
	       &(e->data[e->path2_offset])));
}

static void (*ss_message_log_lut[SS_MESSAGE_TYPE_COUNT])(
	struct ss_message_header *) = {
	[SS_MESSAGE_TYPE_FileSystemEvent] = &ss_fs_event_message_log,
};

static void ss_message_log(struct ss_message_header *message)
{
	time_t ts;

	void (*log_func)(struct ss_message_header *);
	if (!message) {
		return;
	}

	ts = message->timestamp;

	DEBUG(SS_DEBUG_SOCKET_FAIL_LOG_LEVEL,
	      ("message:\n"
	       "\tts: %s",
	       ctime((time_t *)&(ts))));

	if (message->message_type < 0 ||
	    message->message_type >= SS_MESSAGE_TYPE_COUNT) {
		return;
	}
	log_func = ss_message_log_lut[message->message_type];
	if (log_func) {
		(*log_func)(message);
	}
}

static int report_ss_vfs_event(vfs_handle_struct *handle,
			       vfs_op_type op_type,
			       bool success,
			       int saved_errno,
			       int flags,
			       const char *current_path,
			       const char *previous_path)
{
	struct ss_private_data *pd = NULL;
	char *data = NULL;
	size_t data_length = 0;
	struct ss_fs_event_message *message = NULL;
	int ret = 0;

	const char *action = ss_vfs_opname(op_type);

#ifdef DEBUG_ALL_MESSAGES
	DEBUG(0, ("%s: sending message: %s %s\n", __func__, action, current_path));
#endif

	SMB_VFS_HANDLE_GET_DATA(handle, pd, struct ss_private_data, ret = -1; errno = EFAULT; goto out);

	message = (struct ss_fs_event_message *)ss_message_alloc(
		ss_event_log_ctx(),
		SS_MESSAGE_TYPE_FileSystemEvent,
		SS_FS_EVENT_MESSAGE_SIZE(0));
	if (!message) {
		errno = ENOMEM;
		ret = -1;
		goto out;
	}

	message->flags = (uint32_t)flags;
	message->succeeded = success;
	message->_errno = saved_errno;

	message->ip_addr_offset = data_length;
	message->ip_addr_length = strlen(pd->conn_info.ip_address);
	data_length += message->ip_addr_length;
	data = talloc_strndup(ss_event_log_ctx(),
			      pd->conn_info.ip_address,
			      message->ip_addr_length);
	if (!data) {
		errno = ENOMEM;
		ret = -1;
		goto out;
	}

	message->user_name_offset = data_length;
	message->user_name_length = strlen(pd->conn_info.user);
	data_length += message->user_name_length;
	data = talloc_strndup_append(data,
				     pd->conn_info.user,
				     message->user_name_length);
	if (!data) {
		errno = ENOMEM;
		ret = -1;
		goto out;
	}

	message->remote_machine_name_offset = data_length;
	message->remote_machine_name_length = strlen(pd->conn_info.remote_machine_name);
	data_length += message->remote_machine_name_length;
	data = talloc_strndup_append(data,
				     pd->conn_info.remote_machine_name,
				     message->remote_machine_name_length);
	if (!data) {
		errno = ENOMEM;
		ret = -1;
		goto out;
	}

	message->share_name_offset = data_length;
	message->share_name_length = strlen(pd->conn_info.share_name);
	data_length += message->share_name_length;
	data = talloc_strndup_append(data,
				     pd->conn_info.share_name,
				     message->share_name_length);
	if (!data) {
		errno = ENOMEM;
		ret = -1;
		goto out;
	}

	message->action_offset = data_length;
	message->action_length = strlen(action);
	data_length += message->action_length;
	data = talloc_strndup_append(data, action, message->action_length);
	if (!data) {
		errno = ENOMEM;
		ret = -1;
		goto out;
	}

	if (current_path) {
		message->path_offset = data_length;
		message->path_length = strlen(current_path);
		data_length += message->path_length;
		data = talloc_strndup_append(data,
					     current_path,
					     message->path_length);
		if (!data) {
			errno = ENOMEM;
			ret = -1;
			goto out;
		}
	}

	if (previous_path) {
		message->path2_offset = data_length;
		message->path2_length = strlen(previous_path);
		data_length += message->path2_length;
		data = talloc_strndup_append(data,
					     previous_path,
					     message->path2_length);
		if (!data) {
			errno = ENOMEM;
			ret = -1;
			goto out;
		}
	}

	message = (struct ss_fs_event_message *)ss_message_realloc(
		ss_event_log_ctx(),
		(struct ss_message_header *)message,
		SS_FS_EVENT_MESSAGE_SIZE(data_length));
	if (!message) {
		errno = ENOMEM;
		ret = -1;
		goto out;
	}

	memcpy(message->data, data, data_length);

	ret = ss_message_send(handle, (struct ss_message_header *)message);

	if (ret < 0) {
		DEBUG(0,
		      ("%s: Error sending message: %s\n",
		       __func__,
		       strerror(errno)));
		ss_message_log((struct ss_message_header *)message);
	}

out:
	if (ret) {
		DEBUG(0, ("%s: Failed to send message to radar: %s\n", __func__, strerror(errno)));
	}

	TALLOC_FREE(tmp_ss_event_log_ctx);
	return ret;
}

/**
 * Return a string using the ss_event_log_ctx()
 */
static const char *smb_fname_str_event_log(struct connection_struct *conn,
					   const struct smb_filename *smb_fname)
{
	char *fname = NULL;
	NTSTATUS status;

	if (smb_fname == NULL) {
		return "";
	}

	if (smb_fname->base_name[0] != '/') {
		char *abs_name = NULL;
		struct smb_filename *fname_copy =
			cp_smb_filename(ss_event_log_ctx(), smb_fname);
		if (fname_copy == NULL) {
			return "";
		}

		if (!ISDOT(smb_fname->base_name)) {
			abs_name = talloc_asprintf(
				ss_event_log_ctx(),
				"%s/%s",
				conn->cwd_fsp->fsp_name->base_name,
				smb_fname->base_name);
		} else {
			abs_name = talloc_strdup(
				ss_event_log_ctx(),
				conn->cwd_fsp->fsp_name->base_name);
		}
		if (abs_name == NULL) {
			return "";
		}
		fname_copy->base_name = abs_name;
		smb_fname = fname_copy;
	}

	status = get_full_smb_filename(ss_event_log_ctx(), smb_fname, &fname);
	if (!NT_STATUS_IS_OK(status)) {
		return "";
	}
	return fname;
}

static int
ss_connect(vfs_handle_struct *handle, const char *service, const char *user)
{
	int result;
	struct ss_private_data *pd = NULL;

	ss_log_euid();

	result = SMB_VFS_NEXT_CONNECT(handle, service, user);
	if (result < 0) {
		return result;
	}

	pd = talloc_zero(handle, struct ss_private_data);
	if (!pd) {
		DEBUG(0, ("talloc_zero() failed\n"));
		goto err_out;
	}
	if ((pd->conn_info.ip_address = talloc_sub_basic(
		     handle,
		     user,
		     handle->conn->session_info->info->domain_name,
		     "%I")) == NULL) {
		DEBUG(0, ("talloc_sub_basic() for ip_address (%%I) failed\n"));
		goto err_out;
	}
	if ((pd->conn_info.user = talloc_strdup(handle, user)) == NULL) {
		DEBUG(0, ("talloc_strdup() failed\n"));
		goto err_out;
	}
	if ((pd->conn_info.remote_machine_name = get_remote_machine_name()) == NULL) {
		DEBUG(0, ("get_remote_machine_name() failed\n"));
		goto err_out;
	}
	if ((pd->conn_info.share_name = talloc_strdup(handle, service)) ==
	    NULL) {
		DEBUG(0, ("talloc_strdup() failed\n"));
		goto err_out;
	}
	pd->socket_info.socket_path =
		lp_parm_const_string(SNUM(handle->conn),
				     "snapshield",
				     "socket path",
				     SS_DEFAULT_SOCKET_PATH);
	pd->socket_info.socket_connection_fd =
		ss_open_socket_connection(pd->socket_info.socket_path);

	SMB_VFS_HANDLE_SET_DATA(handle,
				pd,
				NULL,
				struct ss_private_data,
				return -1);

	report_ss_vfs_event(handle, SMB_VFS_OP_CONNECT, true, 0, 0, NULL, NULL);

	return result;

err_out:
	SMB_VFS_NEXT_DISCONNECT(handle);
	return -1;
}

static void ss_disconnect(vfs_handle_struct *handle)
{
	/* Perform whatever actions are needed here */
	struct ss_private_data *pd = NULL;

	SMB_VFS_HANDLE_GET_DATA(handle, pd, struct ss_private_data, goto out;);

	// // DEBUG(0, ("connection closed to %s from %s\n", pd->share_name,
	// pd->user));
	report_ss_vfs_event(handle,
			    SMB_VFS_OP_DISCONNECT,
			    true,
			    0,
			    0,
			    NULL,
			    NULL);

	close(pd->socket_info.socket_connection_fd);

out:
	SMB_VFS_NEXT_DISCONNECT(handle);
}

static int ss_openat(struct vfs_handle_struct *handle,
		     const struct files_struct *dirfsp,
		     const struct smb_filename *smb_fname,
		     struct files_struct *fsp,
		     const struct vfs_open_how *how)
{
	int result;
	int saved_errno;

	result = SMB_VFS_NEXT_OPENAT(handle, dirfsp, smb_fname, fsp, how);
	if (result == -1) {
		saved_errno = errno;
	}

	if ((how->flags & O_WRONLY) || (how->flags & O_RDWR)
#if defined(SS_REPORT_OPENAT_READS) && (SS_REPORT_OPENAT_READS == 1)
	    || !(ISDOT(smb_fname->base_name) || ISDOTDOT(smb_fname->base_name))
#endif
	) {
		report_ss_vfs_event(
			handle,
			SMB_VFS_OP_OPENAT,
			result != -1,
			saved_errno,
			how->flags,
			smb_fname_str_event_log(fsp->conn, fsp->fsp_name),
			NULL);
	}

	if (result == -1) {
		errno = saved_errno;
	}
	return result;
}

static int ss_renameat(struct vfs_handle_struct *handle,
		       struct files_struct *srcdir_fsp,
		       const struct smb_filename *smb_fname_src,
		       struct files_struct *dstdir_fsp,
		       const struct smb_filename *smb_fname_dst)
{
	int result;
	int saved_errno;
	struct smb_filename *full_fname_src = NULL;
	struct smb_filename *full_fname_dst = NULL;

	full_fname_src = full_path_from_dirfsp_atname(talloc_tos(),
						      srcdir_fsp,
						      smb_fname_src);
	if (full_fname_src == NULL) {
		errno = ENOMEM;
		return -1;
	}
	full_fname_dst = full_path_from_dirfsp_atname(talloc_tos(),
						      dstdir_fsp,
						      smb_fname_dst);
	if (full_fname_dst == NULL) {
		TALLOC_FREE(full_fname_src);
		errno = ENOMEM;
		return -1;
	}

	result = SMB_VFS_NEXT_RENAMEAT(handle,
				       srcdir_fsp,
				       smb_fname_src,
				       dstdir_fsp,
				       smb_fname_dst);
	if (result == -1) {
		saved_errno = errno;
	}

	report_ss_vfs_event(
		handle,
		SMB_VFS_OP_RENAMEAT,
		result != -1,
		saved_errno,
		0,
		smb_fname_str_event_log(handle->conn, full_fname_dst),
		smb_fname_str_event_log(handle->conn, full_fname_src));

	TALLOC_FREE(full_fname_src);
	TALLOC_FREE(full_fname_dst);

	if (result == -1) {
		errno = saved_errno;
	}
	return result;
}

static int ss_unlinkat(struct vfs_handle_struct *handle,
		       struct files_struct *srcdir_fsp,
		       const struct smb_filename *smb_fname,
		       int flags)
{
	int result;
	int saved_errno;
	struct smb_filename *full_fname = NULL;

	full_fname = full_path_from_dirfsp_atname(talloc_tos(),
						  srcdir_fsp,
						  smb_fname);

	if (full_fname == NULL) {
		errno = ENOMEM;
		return -1;
	}

	result = SMB_VFS_NEXT_UNLINKAT(handle, srcdir_fsp, smb_fname, flags);
	if (result == -1) {
		saved_errno = errno;
	}

	report_ss_vfs_event(handle,
			    SMB_VFS_OP_UNLINKAT,
			    result != -1,
			    saved_errno,
			    0,
			    smb_fname_str_event_log(handle->conn, full_fname),
			    NULL);

	if (result == -1) {
		errno = saved_errno;
	}
	return result;
}

static struct vfs_fn_pointers vfs_snapshield_fns = {
	.connect_fn = ss_connect,
	.disconnect_fn = ss_disconnect,
	.openat_fn = ss_openat,
	.renameat_fn = ss_renameat,
	.unlinkat_fn = ss_unlinkat,
};

static_decl_vfs;
NTSTATUS vfs_snapshield_init(TALLOC_CTX *ctx)
{
	ss_log_euid();
	return smb_register_vfs(SMB_VFS_INTERFACE_VERSION,
				"snapshield",
				&vfs_snapshield_fns);
}
